import json
import threading
from os import getenv as env

from loguru import logger

from common.networking.intercom import Intercom
from common.schnell.workers import Schnell, work
from common.utils.Task import Task
from common.utils.io import getFilesInFileFolder


def convertPayload(workflow):
	convertedWorkflow = list()
	for itemGroup in workflow['tasks']:
		convertedItemGroup = list()
		for item in itemGroup:
			convertedItemGroup.append(Task.fromJSON(item))
		convertedWorkflow.append(convertedItemGroup)

	return [json.loads(k) for k in workflow['arguments']], convertedWorkflow


class Client:
	pool: Schnell = None
	intercom: Intercom = None

	def __init__(self, pool: Schnell):
		self.pool = pool

	def serve(self, allowRemote):
		# Handles the client daemon
		intercom = Intercom("*" if allowRemote else "127.0.0.1", int(env("CLIENT_PORT")), False)

		logger.info("Listening for messages...")
		while True:
			message = intercom.listen()
			logger.info("Received message")
			if not type(message) is dict:
				logger.error("Received invalid message.")
				intercom.ko()
			elif "type" not in message:
				logger.error("Received corrupted message. Throwing away.")
				intercom.ko()
			else:
				# SHUT DOWN
				if message["type"] == "shutdown":
					logger.warning("==> | !!! Shutdown !!! | <==")
					intercom.ok()
					break

				# HEART BEAT
				elif message["type"] == "heartbeat":
					logger.debug("Received heartbeat, sending alive signal.")
					intercom.talk({"type": "alive"})

				# JOBS
				elif message["type"] == "jobs":
					if "jobs" in message:
						logger.info("Received jobs, adding to queue")

						args, tasks = convertPayload(json.loads(message["jobs"]))
						for arg in args:
							data = {"id": arg['id'], 'wargs': arg['data'], "tasks": tasks}
							self.pool.pool.map_async(work, (data,))
						intercom.ok()
					else:
						logger.error("Jobs order is corrupted, ignoring")
						intercom.ko()

				# STATREP
				elif message["type"] == "statrep":
					files = getFilesInFileFolder()
					processPool = self.pool.nWorkers
					intercom.talk({"processPool": processPool, "files": files})

				# THROTTLE
				elif message["type"] == "throttle":
					logger.critical("Received throttle request from server. Stopping all activities for 1 minute.")
					self.pool.pause()
					threading.Timer(180, self.pool.resume).start()
					intercom.ok()

				# UNKNOWN MESSAGE
				else:
					logger.error("Unrecognised message. Ignoring.")
					intercom.ko()

		logger.warning("Stopping client.")
