import json
import threading
from os import getenv as env

import falcon
import requests
from loguru import logger

from common.networking.intercom import Intercom
from common.schnell.workers import Schnell
from server.io.YAMLParser import YAMLParser
# Cleans the context
from server.io.shotgun import Shotgun


def cleanContext(rawContext: dict, clientTasks: list):
	flatten = lambda l: [item for sublist in l for item in sublist]
	items = flatten(clientTasks)
	output = {}

	for item in items:
		itemDependencies = [(v['uuid'], v['argument']) for v in item.dependencies.values()]
		for dependency, argument in itemDependencies:
			if dependency in list(rawContext.keys()):
				output[dependency] = rawContext[dependency]

	return output


# Converts a workflow into a JSON workflow
def toPayload(workflow: list) -> list:
	output = list()

	for itemGroup in workflow:
		convertedItemGroup = list()
		for item in itemGroup:
			convertedItemGroup.append(item.toJSON())
		output.append(convertedItemGroup)

	return output


def compute_response(playbookData):
	logger.info("Received potential YAML file, reading and executing")
	playbook = YAMLParser(playbookData)
	logger.info("YAML was valid and is now ready for dispatch.")
	arguments, clientTasks = playbook.parse()

	return arguments, clientTasks


# Runs a playbook given in argument
def run_playbook(arguments, clientsTasks, clientID):
	serverPayload = requests.get(url="http://" + env('SERVER_WEB_FQDN') + "/api/client/" + clientID,
	                     verify=False, headers={'Accept': 'application/json'})

	if serverPayload.status_code != 200:
		logger.error("Cannot reach mother ship.")
		return

	serverPayload = json.loads(serverPayload.content)
	payload = json.dumps({
		"arguments": [json.dumps(k) for k in serverPayload['parameters']],
		"tasks": toPayload(clientsTasks)
	})

	# Sending first to planisphere
	clientIntercom = Intercom(serverPayload['ip'], int(env("CLIENT_PORT")), True)
	clientIntercom.talk({"type": "jobs", "jobs": payload})
	clientIntercom.listen()

	logger.success("Done, client has been advised")


class ArgumentAPI(object):
	def on_get(self, req: falcon.Request, resp: falcon.Response):
		""" Returns potential arguments for the running of the playbook"""
		data = req.stream.read(req.content_length or 0)
		if len(data) == 0:
			resp.status = falcon.HTTP_422
			resp.body = {"status": "ERROR", "message": "Missing playbook."}
		else:
			# Reading YAML and checking pre-launch requirements
			arguments, _ = compute_response(data)
			resp.status = falcon.HTTP_200
			resp.body = json.dumps(list({argument.params['name'] for argument in arguments}))


class PlaybookAPI(object):
	schnell: Schnell = None

	def setSchnell(self, schnell: Schnell):
		self.schnell = schnell

	def on_get(self, req: falcon.Request, resp: falcon.Response):
		"""Runs the playbook"""
		data = req.stream.read(req.content_length or 0)

		if len(data) == 0:
			resp.status = falcon.HTTP_422
			resp.body = {"status": "ERROR", "message": "Missing playbook."}
		elif "client" not in req.params:
			resp.status = falcon.HTTP_422
			resp.body = {"status": "ERROR", "message": "Missing client."}
		else:
			# Reading YAML and checking pre-launch requirements
			arguments, clientTasks = compute_response(data)

			threading.Thread(target=run_playbook,
			                 args=(arguments, clientTasks, req.params['client'])) \
				.start()
			resp.status = falcon.HTTP_200  # This is the default status
			resp.body = {"status": "OK, enqueued"}

		resp.body = json.dumps(resp.body)
