import argparse

import requests

from dotenv import load_dotenv, find_dotenv
from loguru import logger

from client.client import Client
from common.schnell.workers import Schnell
from os import getenv as env

from server import api

# Environment awareness activation
load_dotenv(find_dotenv())

# Option parsing and checking
parser = argparse.ArgumentParser(description='Flexible and scalable document parser')
group = parser.add_mutually_exclusive_group()
group.add_argument("--server", help="Considers this instance as a server", default=False, action='store_true')
group.add_argument("--client", help="Considers this instance as a client. On by default.", default=True,
                   action='store_true')

parser.add_argument("--remote", help="Allows remote jobs to be sent to this instance. Only used when working in "
                                     "client mode.", default=False, action='store_true')

args = parser.parse_args()

# Identity awareness evaluation
isServer = args.server
allowRemote = args.remote


# Starting program
if __name__ == '__main__':
	logger.info("Inari - Théo Pirkl")
	if not allowRemote:
		logger.warning("You have forbidden outside instances to connect to this instance. \nThat's fine, but you will "
		               "need to start another instance at the same time on this machine with the other mode in order "
		               "for everything to work.")
	if isServer:
		logger.add("run_server.log", format="{time} \t| {process.id} \t| {level}       | {message}", enqueue=True)
		logger.info("Running in server mode. All clients will be controlled from this server.")

		serverPool = Schnell(int(env("SERVER_WORKERS")))
		logger.info("Starting local tasks pool...")
		serverPool.start()

		logger.info("Starting inarid...")
		api.start(serverPool)

		# The API has stopped either because of an error or it stopped gracefuly
		logger.info("Stopped inarid.")

		logger.info("Stopping local tasks pool")
		serverPool.stop()

		logger.info("Stopped all services. Shutting down.")
	else:
		logger.add("run_client.log", format="{time} \t| {process.id} \t| {level}       | {message}", enqueue=True)
		# Specific imports
		from client import HTTPHandler

		logger.info("Running in client mode. All order received by any server will be processed.")
		clientPool = Schnell(int(env("CLIENT_WORKERS")))

		logger.info("Starting local management web server...")
		HTTPHandler.start(int(env("HTTPHANDLER_WEBSERVER_PORT")))

		# Starting workers. Jobs can be executed below this line !!
		logger.info("Starting workers...")
		clientPool.start()

		logger.info("Done. Ready to process orders...")
		client = Client(clientPool)
		client.serve(allowRemote)

		logger.info("Received shutdown signal. Stopping.")
		logger.info("Stopping workers...")
		clientPool.stop()
		# Stopped workers. Jobs won't be executed below this line.

		logger.info("Closing HTTP Handler...")
		requests.get("http://127.0.0.1:{0}/shutdown".format(int(env("HTTPHANDLER_WEBSERVER_PORT"))))

	# Everything has been stopped.
	logger.success("All services have been shut down. Good night.")
