<?php

namespace App\Http\Controllers\API;

use App\Http\Controllers\Controller;
use App\Http\Requests\WorkflowJSONRequest;
use App\Http\Requests\WorkflowRequest;
use App\Models\Client;
use App\Models\Job;
use App\Models\Workflow;
use Illuminate\Contracts\Filesystem\FileNotFoundException;
use Illuminate\Http\Request;
use Illuminate\Http\Response;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Http;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Storage;
use Illuminate\Validation\ValidationException;

class WorkflowController extends Controller {
    function __construct(){
        $this->middleware('auth:api');
    }

    /**
     * Display a listing of the resource.
     *
     * @return Response
     */
    public function index(){
        return Auth::user()->workflows;
    }

    /**
     * Shows a selected resource.
     * @param Workflow $workflow
     * @return Workflow
     */
    public function show(Workflow $workflow){
        return $workflow;
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param WorkflowRequest $request
     */
    public function store(WorkflowRequest $request){
        $w = Workflow::create(['name' => $request->name, 'user_id' => $request->user()->id]);
        Storage::disk('public')->put('workflow-'.$w->id.'.json', collect([]));
    }

    /**
     * Update the specified resource in storage.
     *
     * @param WorkflowRequest $request
     * @param Workflow $workflow
     * @return void
     */
    public function update(WorkflowRequest $request, Workflow $workflow){
        $workflow->name = $request->name;
        $workflow->save();
    }

    /**
     * Remove the specified resource from storage.
     *
     * @param Workflow $workflow
     * @return void
     * @throws \Exception
     */
    public function destroy(Workflow $workflow){
        $workflow->modules()->detach();
        $workflow->delete();
    }

    /**
     * Stores the JSON form of a workflow in a file.
     * @param WorkflowJSONRequest $request
     * @param Workflow $workflow
     */
    public function storeJSON(WorkflowJSONRequest $request, Workflow $workflow){
        Storage::disk('public')->put('workflow-'.$workflow->id.'.json', $request->data);
    }

    /**
     * Gets a JSON representation of the workflow
     * @param Workflow $workflow
     * @return string
     * @throws \Illuminate\Contracts\Filesystem\FileNotFoundException
     */
    public function getJSON(Workflow $workflow){
        return Storage::disk('public')->get('workflow-'.$workflow->id.'.json');
    }

    /**
     * Gets the arguments for a workflow
     * @param Workflow $workflow
     */
    public function getArguments(Workflow $workflow){
        try {
            $file = Storage::disk('local')->get('builds/build-'.$workflow->id.'.yaml');
            return Http::withOptions(['body' => $file])->get("http://127.0.0.1:8080/playbookArgs")->json();
        } catch (FileNotFoundException $e) {
            throw ValidationException::withMessages(['workflow' => 'You must first build your workflow.']);
        }
    }

    /**
     * Gets the most available client
     * @return Client
     */
    private function getBestClient(){
        return Client::with('jobs')->get()
            ->sort(function($clientA, $clientB){
                return $clientA->jobs->count() < $clientB->jobs->count() ? -1 : 1;
            })->first();
    }

    /**
     * Runs a playbook.
     * @param Request $request
     * @param Workflow $workflow
     * @return array
     * @throws ValidationException
     */
    public function run(Request $request, Workflow $workflow){
        // Choosing least loaded client
        $nextClient = $this->getBestClient();

        Job::create([
            'client_id'     => $nextClient->id,
            'workflow_id'   => $workflow->id,
            'data'          => $request->params
        ]);

        // Formatting for inari server
        $params = ['client' => $nextClient->id];

        return $this->talkToClient($workflow, $params);
    }

    /**
     * Launches many requests in the gueule of one client
     * @param Request $request
     * @param Workflow $workflow
     * @return void
     */
    public function bulk(Request $request, Workflow $workflow){
        // Mapping arguments for DB insertion
        $params = collect($request->params)->map(function($entry) use ($request, $workflow){
            return [
                'data'          => $entry,
                'client_id'     => $request->client,
                'workflow_id'   => $workflow->id
            ];
        })->toArray();


        Client::findOrFail($request->client)->jobs()->createMany($params); // Hell yes
    }

    /**
     * @param Request $request
     * @param Workflow $workflow
     */
    public function runBulk(Request $request, Workflow $workflow){
        foreach ($request->clients as $client) {
            try {
                $this->talkToClient($workflow, ['client' => $client]);
            } catch (ValidationException $e) {
                Log::error("Cannot talk to client ID'd ".$client);
            }
        }
    }

    /**
     * @param Workflow $workflow
     * @param array $params
     * @return array
     * @throws ValidationException
     */
    public function talkToClient(Workflow $workflow, array $params){
        try {
            $file = Storage::disk('local')->get('builds/build-'.$workflow->id.'.yaml');
            return Http::withOptions(['body' => $file])
                ->get("http://127.0.0.1:8080/playbook", $params)->json();
        } catch (FileNotFoundException $e) {
            throw ValidationException::withMessages(['workflow' => 'You must first build your workflow.']);
        }
    }
}
