<?php

namespace App\Jobs;

use App\Models\Build;
use App\Models\Module;
use Illuminate\Bus\Queueable;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Queue\SerializesModels;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Storage;

class BuildPlaybook implements ShouldQueue {
    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    /**
     * @var Build
     */
    private $build;

    /**
     * Create a new job instance.
     *
     * @param Build $build
     */
    public function __construct(Build $build){
        $this->build = $build;
    }

    /**
     * Gets the children of at least one element
     * @param $elements Collection
     * @param $file Collection The JSON representation of the file
     * @return mixed
     */
    private function children($elements, $file){
        $ids = $elements->map(function($item){
            return collect($item->connections->outputs)->map(function($connection){
                return $connection->module;
            })->values();
        })->flatten();
        if ($ids->isNotEmpty()){
            return $file->filter(function($moduleInfo) use ($ids){
                return $ids->contains($moduleInfo->module->uuid);
            });
        } else {
            return collect();
        }
    }

    /**
     * Gets the parents of at least one element
     * @param $elements
     * @return mixed
     */
    private function parents($elements, $file){
        $ids = $elements->map(function($item){
            return collect($item->connections->inputs)->map(function($connection){
                return $connection->module;
            })->values();
        })->flatten();
        if ($ids->isNotEmpty()){
            return $file->filter(function($moduleInfo) use ($ids){
                return $ids->contains($moduleInfo->module->uuid);
            });
        } else {
            return collect();
        }
    }

    /**
     * Returns a module info by its UUID
     * @param $uuid string The UUID
     * @param $file Collection The global file info
     * @return mixed A module info
     */
    private function findByUUID($uuid, $file){
        return $file->first(function($moduleInfo) use ($uuid){
            return $moduleInfo->module->uuid = $uuid;
        });
    }

    /**
     * Creates a map order
     * @param $output Collection
     * @param $elements Collection
     * @param $file Collection
     */
    private function iterate($output, $elements, $file){
        $children = $this->parents($elements, $file);
        if ($children->isNotEmpty()){
            $output->get('actions')->push($children->toArray());
            $this->iterate($output, $children, $file);
        }
    }

    /**
     * Cleans the map order to remove any duplicates
     * @param $elements Collection The map order
     * @return Collection The cleaned map order
     */
    private function clean($elements){
        $output = $elements->toBase();
        $output['actions'] = $output->get('actions')->reverse();
        $visitedUUIDs = collect();
        // Removing duplicated only when they're seen more than once
        $filteredActions = $output->get('actions')->map(function($group) use ($visitedUUIDs){
            return collect($group)->filter(function($module) use ($visitedUUIDs){
                if ($visitedUUIDs->contains($module->module->uuid)){
                    return false;
                } else {
                    $visitedUUIDs->push($module->module->uuid);
                    return true;
                }
            });
        });
        // Reversing again (we're done)
        $output->put('actions', $filteredActions);
        // All good :-)
        return $output;
    }

    /**
     * Execute the job.
     *
     * @throws \Illuminate\Contracts\Filesystem\FileNotFoundException
     */
    public function handle(){
        // Let's mark the build as "WORKING"
        $this->build->state = 1;
        $this->build->save();

        // Working
        $file = collect(json_decode(Storage::disk('public')->get('workflow-'.$this->build->workflow->id.'.json')));
        $output = collect(['actions' => collect()]);

        // Last elements are IMMEDIATELY put inside the output.
        $lastElements = $file->filter(function($module){
            return collect($module->connections->outputs)->isEmpty();
        });
        $output->get('actions')->push($lastElements->toArray());

        // Now, we iterate on each child.
        $this->iterate($output, $lastElements, $file);

        // We now need to clean the data (remove duplicates)
        $output = $this->clean($output);

        // The output is almost ready ! We only need to convert the module info into a proper module
        $newActions = collect();
        $output->get('actions')->each(function($group) use ($file, $newActions) {
            $actionGroup = collect($group)->mapWithKeys(function($item) use ($file){
                $module = Module::findOrFail($item->module->id);
                $parents = $this->parents(collect([$item]), $file);
                return [$item->module->uuid => $module->toYAML($item)];
            })->toArray();
            $newActions->push($actionGroup);
        });
        $output['actions'] = $newActions;


        // Let's get rid of the laravel objects
        $output = ['actions' => $output['actions']->toArray()];

        // We're ready.
        $output = yaml_emit($output);
        Storage::disk('local')->put('builds/build-'.$this->build->workflow->id.'.yaml', $output);

        // Done. Marking as done in DB.
        $this->build->state = 2;
        $this->build->save();
    }
}
